/*
 * Copyright (c) 2011 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Thu Apr  7 2011
 */
#include <err.h>

#include "dc1394_common.h"


/* --- local data ---------------------------------------------------------- */

dc1394_t *dc1394h;		/**< general libdc134 handle */

static int	viamgen1394_fwfeature(viam_filtertype_t filter,
			unsigned int *fwfeat);

static int	viamgen1394_hw_geo(uint32_t w, uint32_t h, viam_hwformat_t fmt,
			dc1394video_mode_t *hwgeo);
static int	viamgen1394_hw_geo_7(viam_handle_t vh, dc1394camera_t *handle,
			uint32_t w, uint32_t h, viam_hwformat_t fmt,
			dc1394video_mode_t *hwgeo, dc1394color_coding_t *color);

static int	viamgen1394_hwfps(double fps, dc1394framerate_t *hwfps);
static int	viamgen1394_hwtrigger(viam_hwtrigger_t swtrig,
			dc1394trigger_mode_t *hwtrig,
			dc1394trigger_polarity_t *pol);


/* --- viam1394_init ------------------------------------------------------- */

int
viamgen1394_init(viam_handle_t vh)
{
  dc1394h = dc1394_new();
  if (!dc1394h) return viam_seterrno(vh, Eviamlib_FWBUS);

  return 0;
}


/* --- viam1394_fini ------------------------------------------------------- */

int
viamgen1394_fini(viam_handle_t vh)
{
  if (dc1394h) dc1394_free(dc1394h);

  return 0;
}


/* --- viam1394_hardware_list ---------------------------------------- */

/** \brief Print available camera devices on all buses
 *
 * \param[in] vh	Library handle.
 * \param[in] print	Callback function displaying camera info.
 *
 * \return 0 on success.
 */

int
viamgen1394_hardware_list(viam_handle_t vh, void (*print)(dc1394camera_t *))
{
  dc1394camera_list_t *clist;
  dc1394camera_t *camera;
  dc1394error_t s;
  int i;

  s = dc1394_camera_enumerate(dc1394h, &clist);
  if (s != DC1394_SUCCESS)
    return viam_seterrno(vh, Eviamlib_NO_CAMERA);

  /* iterate over all cameras */
  for(i=0; i<clist->num; i++) {
    camera = dc1394_camera_new_unit(
      dc1394h, clist->ids[i].guid, clist->ids[i].unit);
    if (!camera) continue;

    print(camera);
    dc1394_camera_free(camera);
  }
  dc1394_camera_free_list(clist);

  return 0;
}


/* --- viamgen1394_hardware_attach ----------------------------------------- */

/** \brief Allocate camera ports and handles
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Camera pretty name.
 * \param[in] uid	Firewire uid.
 *
 * \return a pointer on camera handle on success, or an error.
 */

viamgen1394_camera_t
viamgen1394_hardware_attach(viam_handle_t vh, const char *name, uint64_t uid)
{
  viamgen1394_camera_t c;
  int i;

  dc1394camera_list_t *clist;
  dc1394error_t s;

  /* get all cameras */
  s = dc1394_camera_enumerate(dc1394h, &clist);
  if (s != DC1394_SUCCESS) {
    viam_seterrno(vh, Eviamlib_NO_CAMERA);
    return NULL;
  }

  /* scan all cameras and assign handle if one camera matches */
  for(i=0, c=NULL; i<clist->num; i++) {
    if (clist->ids[i].guid != uid) continue;

    /* register camera */
    c = malloc(sizeof(*c));
    if (!c) {
      warnx("unable to allocate memory for camera %d", i);
      break;
    }
    c->handle = dc1394_camera_new_unit(
      dc1394h, clist->ids[i].guid, clist->ids[i].unit);
    if (!c->handle) {
      warnx("unable to attach to camera `%s'", name);
      free(c);
      c = NULL;
      break;
    }

    s = dc1394_feature_get_all(c->handle, &c->featset);
    if (s != DC1394_SUCCESS) {
      warnx("unable to get camera `%s' features", name);
      dc1394_camera_free(c->handle);
      free(c);
      c = NULL;
      break;
    }

    c->xmit = 0;
    s = dc1394_video_set_transmission(c->handle, DC1394_OFF);
    if (s != DC1394_SUCCESS) viam_seterrno(vh, Eviamlib_FWBUS);

    warnx("camera `%s' is %s, %s", name, c->handle->model, c->handle->vendor);
    break;
  }

  dc1394_camera_free_list(clist);

  if (c == NULL) viam_seterrno(vh, Eviamlib_NO_SUCH_CAMERA);
  return c;
}


/* --- viamgen1394_hardware_detach ----------------------------------------- */

/** \brief Cleanup dc1394 handle
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 */

void
viamgen1394_hardware_detach(viam_handle_t vh, viamgen1394_camera_t hw)
{
  if (hw) {
    dc1394_camera_free(hw->handle);
    free(hw);
  }
}


/* --- viam1394_camera_configure ------------------------------------- */

/** \brief Setup camera with current configuration parameters.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 *
 * \return 0 on success
 */

int
viamgen1394_camera_configure(viam_handle_t vh, viamgen1394_camera_t hw,
                             viam_hwcrop_t crop, uint32_t w, uint32_t h,
                             viam_hwformat_t fmt, double fps,
                             viam_hwtrigger_t trigger)
{
  dc1394video_mode_t fwmode;
  dc1394format7mode_t f7_mode = { .present = 1 };
  dc1394color_coding_t f7_color;
  dc1394framerate_t fwfps;
  dc1394trigger_mode_t fwtrig;
  dc1394trigger_polarity_t fwpol;
  dc1394bool_t p;
  dc1394error_t s;
  uint64_t image_size;
  uint32_t unit_bytes, max_bytes;
  int x, y;
  int psize, npkt;

  /* configure hardware mode */
  if (crop == VIAM_HW_CROP) {
    if (viamgen1394_hw_geo_7(vh, hw->handle, w, h, fmt, &fwmode, &f7_color))
      return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
  } else {
    if (viamgen1394_hw_geo(w, h, fmt, &fwmode))
      return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
  }

  s = dc1394_video_set_mode(hw->handle, fwmode);
  if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);

  /* configure fps */
  if (crop == VIAM_HW_CROP) {
    if (fps == 0.0)
      psize = DC1394_USE_MAX_AVAIL;
    else {
      s = dc1394_format7_get_total_bytes(hw->handle, fwmode, &image_size);
      if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);

      dc1394_format7_get_packet_parameters(
	hw->handle, fwmode, &unit_bytes, &max_bytes);
      if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);

      npkt = 0.8*8000/fps + 0.5; /* 80% of 8kHz */
      psize = (image_size + npkt - 1) / npkt;
      psize = unit_bytes * ((psize + unit_bytes - 1)/unit_bytes);
      if (psize > max_bytes) psize = max_bytes;
    }
  } else {
    if (viamgen1394_hwfps(fps, &fwfps))
      return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);

    s = dc1394_video_set_framerate(hw->handle, fwfps);
    if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);
  }

  /* configure roi (format7 only) */
  if (crop == VIAM_HW_CROP) {
    s = dc1394_format7_get_mode_info(hw->handle, fwmode, &f7_mode);
    if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);

    if (w % f7_mode.unit_size_x)
      w += f7_mode.unit_size_x - (w % f7_mode.unit_size_x);
    if (h % f7_mode.unit_size_y)
      h += f7_mode.unit_size_y - (h % f7_mode.unit_size_y);

    if (f7_mode.unit_pos_x == 0)
      x = 0;
    else {
      x = (f7_mode.max_size_x - w)/2;
      x -= x % f7_mode.unit_pos_x;
    }

    if (f7_mode.unit_pos_y == 0)
      y = 0;
    else {
      y = (f7_mode.max_size_y - h)/2;
      y -= y % f7_mode.unit_pos_y;
    }

    /* first set size, then use set_roi, to avoid passing through an
     * invalid color_coding + old w,h */
    s = dc1394_format7_set_image_size(hw->handle, fwmode, w, h);
    s = dc1394_format7_set_roi(
      hw->handle, fwmode, f7_color, psize, x, y, w, h);
    if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);
  }

  /* configure external trigger */
  p = hw->featset.
    feature[DC1394_FEATURE_TRIGGER-DC1394_FEATURE_MIN].available;

  switch(trigger) {
    case VIAM_HWTRIGGER_INTERNAL:
      if (p == DC1394_TRUE)
	dc1394_external_trigger_set_power(hw->handle, DC1394_OFF);
      break;

    default:
      if (p != DC1394_TRUE)
	return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);

      if (viamgen1394_hwtrigger(trigger, &fwtrig, &fwpol))
	return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
      s = dc1394_external_trigger_set_mode(hw->handle, fwtrig);
      if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);

      s = dc1394_external_trigger_has_polarity(hw->handle, &p);
      if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);
      if (p == DC1394_TRUE) {
	s = dc1394_external_trigger_set_polarity(hw->handle, fwpol);
	if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);
      } else {
	dc1394trigger_polarity_t cpol;

	s = dc1394_external_trigger_get_polarity(hw->handle, &cpol);
	if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);
	if ((cpol & 1) != (fwpol & 1))
	  return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
      }

      s = dc1394_external_trigger_set_power(hw->handle, DC1394_ON);
      if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);
      break;
  }

  /* refresh features because of mode change */
  s = dc1394_feature_get_all(hw->handle, &hw->featset);
  if (s != DC1394_SUCCESS) {
    warnx("unable to get camera `%s %s' features",
          hw->handle->vendor, hw->handle->model);
    return viam_seterrno(vh, Eviamlib_FWBUS);
  }

  /* setup dma */
  s = dc1394_capture_setup(hw->handle, 2/*dma buffers*/,
			   DC1394_CAPTURE_FLAGS_DEFAULT);
  if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);

  return 0;
}


/* --- viamgen1394_camera_unconfigure -------------------------------------- */

/** \brief Reset bank configuration
 *
 * \param[in] vh	Library handle.
 * \param[in] hw	Camera handle.
 *
 * \return 0 on success
 */

int
viamgen1394_camera_unconfigure(viam_handle_t vh, viamgen1394_camera_t hw)
{
  dc1394error_t s;

  s = dc1394_capture_stop(hw->handle);
  if (s != DC1394_SUCCESS) return Eviamlib_FWBUS;

  return 0;
}


/* --- viamgen1394_datatransmit -------------------------------------------- */

int
viamgen1394_datatransmit(viam_handle_t vh, viamgen1394_camera_t hw,
                         viam_onoff_t on)
{
  dc1394error_t s = DC1394_SUCCESS;

  switch(on) {
    case VIAM_ON:
      if (!hw->xmit) {
        s = dc1394_video_set_transmission(hw->handle, DC1394_ON);
        /* flush old data */
        s = viamgen1394_dataflush(vh, hw);
      }
      hw->xmit++;
      break;

    case VIAM_OFF:
      if (hw->xmit) hw->xmit--;
      if (!hw->xmit) {
        s = dc1394_video_set_transmission(hw->handle, DC1394_OFF);
        if (s != DC1394_SUCCESS) break;

        /* flush old data */
        s = viamgen1394_dataflush(vh, hw);
      }
      break;
  }
  if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);
  return 0;
}


/* --- viamgen1394_dataflush ----------------------------------------------- */

int
viamgen1394_dataflush(viam_handle_t vh, viamgen1394_camera_t hw)
{
  dc1394error_t s;

  /* flush DMA buffer to skip old data */
  while(1) {
    s = dc1394_capture_dequeue(hw->handle,
			       DC1394_CAPTURE_POLICY_POLL, &hw->frame);
    if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);
    if (!hw->frame) /* no more old frames */ break;

    s = dc1394_capture_enqueue(hw->handle, hw->frame);
    if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);
  }

  return 0;
}


/* --- viamgen1394_capture ------------------------------------------------- */

int
viamgen1394_capture(viam_handle_t vh, viamgen1394_camera_t hw)
{
  dc1394error_t s;
  s = dc1394_capture_dequeue(hw->handle,
			     DC1394_CAPTURE_POLICY_WAIT, &hw->frame);
  if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);

  return 0;
}


/* --- viamgen1394_capturedata --------------------------------------------- */

char *
viamgen1394_capturedata(viam_handle_t vh, viamgen1394_camera_t hw,
                        size_t *bytes, size_t *rowstride)
{
  if (bytes) *bytes = hw->frame->image_bytes;
  if (rowstride) *rowstride = hw->frame->stride;
  return (char *)hw->frame->image;
}


/* --- viamgen1394_capturetimestamp ---------------------------------------- */

struct timeval
viamgen1394_capturetimestamp(viam_handle_t vh, viamgen1394_camera_t hw)
{
  struct timeval t = {0, 0};

  t.tv_sec = hw->frame->timestamp / 1000000;
  t.tv_usec = hw->frame->timestamp % 1000000;
  return t;
}


/* --- viamgen1394_capturedone --------------------------------------------- */

int
viamgen1394_capturedone(viam_handle_t vh, viamgen1394_camera_t hw)
{
  dc1394error_t s;

  s = dc1394_capture_enqueue(hw->handle, hw->frame);
  if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);

  return 0;
}


/* --- viamgen1394_queryhwfeature ------------------------------------------ */

int
viamgen1394_queryhwfeature(viam_handle_t vh, viamgen1394_camera_t hw,
                           viam_filtertype_t filter,
                           unsigned int *methods, unsigned int *aumodes)
{
  dc1394feature_info_t *info;
  unsigned int feature;
  int i, s;

  if (methods) *methods = 0;
  if (aumodes) *aumodes = 0;

  s = viamgen1394_fwfeature(filter, &feature);
  if (s) return 0;

  info = &hw->featset.feature[feature - DC1394_FEATURE_MIN];
  if (!info->available) return 0;
  if (info->on_off_capable && methods)
    *methods |= VIAM_FILTER_OFF;

  for(i=0;i<info->modes.num;i++) switch (info->modes.modes[i]) {
    case DC1394_FEATURE_MODE_MANUAL:
      if (methods) *methods |= VIAM_FILTER_HARDWARE;
      if (aumodes) *aumodes |= VIAM_FILTER_MANUAL;
      break;

    case DC1394_FEATURE_MODE_AUTO:
      if (aumodes) *aumodes |= VIAM_FILTER_HARDWARE_AUTO;
      break;

    case DC1394_FEATURE_MODE_ONE_PUSH_AUTO:
      if (aumodes) *aumodes |= VIAM_FILTER_HARDWARE_ONE_PUSH;
      break;
  }

  return 0;
}


/* --- viamgen1394_sethwfeaturemode ---------------------------------------- */

/** \brief Set hardware feature mode for a camera.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 * \param[in] filter	Feature type.
 * \param[in] method	Hardware mode.
 * \param[in] aumode	Auto mode.
 *
 * \return 0 on success.
 */

int
viamgen1394_sethwfeaturemode(viam_handle_t vh, viamgen1394_camera_t hw,
                             viam_filtertype_t filter,
                             viam_filtermethod_t method,
                             viam_filterautomode_t aumode)
{
  unsigned int feature;
  dc1394feature_info_t *info;
  dc1394error_t s;
  dc1394bool_t b;

  s = viamgen1394_fwfeature(filter, &feature);
  if (s) switch(method) {
    case VIAM_FILTER_SOFTWARE:
    case VIAM_FILTER_OFF:
      return 0;

    default: return viam_seterrno(vh, s);
  }
  info = &hw->featset.feature[feature - DC1394_FEATURE_MIN];

  switch(method) {
    case VIAM_FILTER_SOFTWARE:
    case VIAM_FILTER_OFF: {
      /* switch feature off if possible */
      s = dc1394_feature_is_switchable(hw->handle, feature, &b);
      if (s != DC1394_SUCCESS) break;

      if (b == DC1394_TRUE) {
	s = dc1394_feature_set_power(hw->handle, feature, DC1394_OFF);
	break;
      }

      /* try to set it at least to manual mode */
      s = dc1394_feature_set_mode(
	hw->handle, feature, DC1394_FEATURE_MODE_MANUAL);
      break;
    }

    case VIAM_FILTER_HARDWARE:
      /* switch feature on if possible */
      s = dc1394_feature_is_switchable(hw->handle, feature, &b);
      if (s == DC1394_SUCCESS && b == DC1394_TRUE) {
        s = dc1394_feature_set_power(hw->handle, feature, DC1394_ON);
        if (s != DC1394_SUCCESS) break;
      }
      /* switch feature to 'absolute' mode if possible */
      if (info->absolute_capable == DC1394_TRUE) {
        s = dc1394_feature_set_absolute_control(
          hw->handle, feature, DC1394_ON);
        info->abs_control = (s == DC1394_SUCCESS)?DC1394_ON:DC1394_OFF;
      } else
        info->abs_control = DC1394_OFF;

      switch(aumode) {
        case VIAM_FILTER_MANUAL:
	case VIAM_FILTER_SOFTWARE_AUTO:
	case VIAM_FILTER_SOFTWARE_ONE_PUSH:
	  s = dc1394_feature_set_mode(
	    hw->handle, feature, DC1394_FEATURE_MODE_MANUAL);
	  break;

	case VIAM_FILTER_HARDWARE_AUTO:
	  s = dc1394_feature_set_mode(
	    hw->handle, feature, DC1394_FEATURE_MODE_AUTO);
	  break;

	case VIAM_FILTER_HARDWARE_ONE_PUSH:
	  s = dc1394_feature_set_mode(
	    hw->handle, feature, DC1394_FEATURE_MODE_ONE_PUSH_AUTO);
	  break;

	case VIAM_FILTER_AUTOINVALID:
	  s = DC1394_FAILURE;
	  break;
      }
      break;

    case VIAM_FILTER_INVALID:
      return viam_seterrno(vh, Eviamlib_NOT_IMPLEMENTED);
  }
  if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);

  return 0;
}


/* --- viamgen1394_sethwfeaturevalue --------------------------------------- */

/** \brief Set harware feature value for a camera.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 * \param[in] filter	Feature type.
 * \param[in] args	Feature value.
 *
 * \return 0 on success.
 */

int
viamgen1394_sethwfeaturevalue(viam_handle_t vh, viamgen1394_camera_t hw,
                              viam_filtertype_t filter,
                              const viam_filterargs_obj args)
{
  unsigned int feature;
  dc1394feature_info_t *info;
  dc1394error_t s;

  s = viamgen1394_fwfeature(filter, &feature);
  if (s) return viam_seterrno(vh, s);
  info = &hw->featset.feature[feature - DC1394_FEATURE_MIN];

  switch(feature) {
    case DC1394_FEATURE_WHITE_BALANCE:
      if (args->vkind == VIAM_VALUE_PERCENT) {
	s = dc1394_feature_whitebalance_set_value(
	  hw->handle,
	  args->blue*info->max + (1.0 - args->blue)*info->min,
	  args->red*info->max + (1.0 - args->red)*info->min);
      } else {
	s = 1; /* error */
      }
      break;

    default:
      if (args->vkind == VIAM_VALUE_PERCENT) {
	if (info->abs_control == DC1394_ON) {
	  s = dc1394_feature_set_absolute_value(
	    hw->handle, feature,
	    args->value*info->abs_max + (1.0 - args->value)*info->abs_min);
	} else {
	  s = dc1394_feature_set_value(
	    hw->handle, feature,
	    args->value*info->max + (1.0 - args->value)*info->min);
	}
      } else {
	if (info->abs_control == DC1394_ON) {
	  if (args->value < info->abs_min) args->value = info->abs_min;
	  if (args->value > info->abs_max) args->value = info->abs_max;
	  s = dc1394_feature_set_absolute_value(
	    hw->handle, feature, args->value);
	} else {
	  if (args->value < info->min) args->value = info->min;
	  if (args->value > info->max) args->value = info->max;
	  s = dc1394_feature_set_value(hw->handle, feature, args->value);
	}
      }
      break;
  }
  if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);

  return 0;
}


/* --- viam1394_fwfeature -------------------------------------------- */

static const struct {
  unsigned int supported;
  unsigned int fwfeat;
} featuremap[] = {
#ifdef VIAM_OPENCV_SUPPORT
  [VIAM_FILTER_WHITE_BALANCE] =	{ 1, DC1394_FEATURE_WHITE_BALANCE },
#endif
  [VIAM_FILTER_BRIGHTNESS] =	{ 1, DC1394_FEATURE_BRIGHTNESS },
  [VIAM_FILTER_EXPOSURE] =	{ 1, DC1394_FEATURE_EXPOSURE },
  [VIAM_FILTER_SHARPNESS] =	{ 1, DC1394_FEATURE_SHARPNESS },
  [VIAM_FILTER_GAIN] =		{ 1, DC1394_FEATURE_GAIN },
  [VIAM_FILTER_SHUTTER] =	{ 1, DC1394_FEATURE_SHUTTER },
  [VIAM_FILTER_HUE] =		{ 1, DC1394_FEATURE_HUE },
  [VIAM_FILTER_SATURATION] =	{ 1, DC1394_FEATURE_SATURATION },
  [VIAM_FILTER_GAMMA] =		{ 1, DC1394_FEATURE_GAMMA },
  [VIAM_FILTER_TEMPERATURE] =	{ 1, DC1394_FEATURE_TEMPERATURE },
  [VIAM_FILTER_ZOOM] =		{ 1, DC1394_FEATURE_ZOOM },
  [VIAM_FILTER_FOCUS] =		{ 1, DC1394_FEATURE_FOCUS },
  [VIAM_FILTER_PAN] =		{ 1, DC1394_FEATURE_PAN },
  [VIAM_FILTER_TILT] =		{ 1, DC1394_FEATURE_TILT }
};

static int
viamgen1394_fwfeature(viam_filtertype_t filter, unsigned int *fwfeat)
{
  if (!featuremap[filter].supported) return Eviamlib_NOT_IMPLEMENTED;
  *fwfeat = featuremap[filter].fwfeat;
  return 0;
}


/* --- viamgen1394_hw_geo -------------------------------------------------- */

static const struct {
  uint32_t w, h;
  viam_hwformat_t fmt;
  dc1394video_mode_t hw;
} geomap[] = {
  { 160, 120,	VIAM_HWFMT_YUV444,	DC1394_VIDEO_MODE_160x120_YUV444 },
  { 320, 240,	VIAM_HWFMT_YUV422_UYVY,	DC1394_VIDEO_MODE_320x240_YUV422 },
  { 640, 480,	VIAM_HWFMT_YUV411,	DC1394_VIDEO_MODE_640x480_YUV411 },
  { 640, 480,	VIAM_HWFMT_YUV422_UYVY,	DC1394_VIDEO_MODE_640x480_YUV422 },
  { 640, 480,	VIAM_HWFMT_RGB888,	DC1394_VIDEO_MODE_640x480_RGB8 },
  { 640, 480,	VIAM_HWFMT_MONO8,	DC1394_VIDEO_MODE_640x480_MONO8 },
  { 640, 480,	VIAM_HWFMT_MONO16,	DC1394_VIDEO_MODE_640x480_MONO16 },
  { 800, 600,	VIAM_HWFMT_YUV422_UYVY,	DC1394_VIDEO_MODE_800x600_YUV422 },
  { 800, 600,	VIAM_HWFMT_RGB888,	DC1394_VIDEO_MODE_800x600_RGB8 },
  { 800, 600,	VIAM_HWFMT_MONO8,	DC1394_VIDEO_MODE_800x600_MONO8 },
  { 1024, 768,	VIAM_HWFMT_YUV422_UYVY,	DC1394_VIDEO_MODE_1024x768_YUV422 },
  { 1024, 768,	VIAM_HWFMT_RGB888,	DC1394_VIDEO_MODE_1024x768_RGB8 },
  { 1024, 768,	VIAM_HWFMT_MONO8,	DC1394_VIDEO_MODE_1024x768_MONO8 },
  { 800, 600,	VIAM_HWFMT_MONO16,	DC1394_VIDEO_MODE_800x600_MONO16 },
  { 1024, 768,	VIAM_HWFMT_MONO16,	DC1394_VIDEO_MODE_1024x768_MONO16 },
  { 1280, 960,	VIAM_HWFMT_YUV422_UYVY,	DC1394_VIDEO_MODE_1280x960_YUV422 },
  { 1280, 960,	VIAM_HWFMT_RGB888,	DC1394_VIDEO_MODE_1280x960_RGB8 },
  { 1280, 960,	VIAM_HWFMT_MONO8,	DC1394_VIDEO_MODE_1280x960_MONO8 },
  { 1600, 1200,	VIAM_HWFMT_YUV422_UYVY,	DC1394_VIDEO_MODE_1600x1200_YUV422 },
  { 1600, 1200,	VIAM_HWFMT_RGB888,	DC1394_VIDEO_MODE_1600x1200_RGB8 },
  { 1600, 1200,	VIAM_HWFMT_MONO8,	DC1394_VIDEO_MODE_1600x1200_MONO8 },
  { 1280, 960,	VIAM_HWFMT_MONO16,	DC1394_VIDEO_MODE_1280x960_MONO16 },
  { 1600, 1200,	VIAM_HWFMT_MONO16,	DC1394_VIDEO_MODE_1600x1200_MONO16 }
};

static int
viamgen1394_hw_geo(uint32_t w, uint32_t h, viam_hwformat_t fmt,
                   dc1394video_mode_t *hwgeo)
{
  int i;

  for(i=0; i<sizeof(geomap)/sizeof(geomap[0]); i++)
    if (geomap[i].w == w && geomap[i].h == h && geomap[i].fmt == fmt) {
      *hwgeo = geomap[i].hw;
      return 0;
    }

  return Eviamlib_UNSUPPORTED_HWMODE;
}

static int
viamgen1394_hw_geo_7(viam_handle_t vh, dc1394camera_t *handle,
                     uint32_t w, uint32_t h, viam_hwformat_t fmt,
                     dc1394video_mode_t *hwgeo, dc1394color_coding_t *color)
{
  dc1394video_modes_t modes;
  dc1394format7mode_t f7_mode = { .present = 1 };
  dc1394error_t s;
  int minw = 32768, minh = 32768;
  int i, j;

  *hwgeo = -1;
  s = dc1394_video_get_supported_modes(handle, &modes);
  if (s != DC1394_SUCCESS) return viam_seterrno(vh, Eviamlib_FWBUS);
  for (i=0; i<modes.num; i++) {
    if (!dc1394_is_video_mode_scalable(modes.modes[i])) continue;

    s = dc1394_format7_get_mode_info(handle, modes.modes[i], &f7_mode);
    if (s != DC1394_SUCCESS) continue;
    if (f7_mode.max_size_x < w || f7_mode.max_size_y < h) continue;

    /* map image format */
    for(j=0; j<f7_mode.color_codings.num; j++) {
      switch(f7_mode.color_codings.codings[j]) {
	case DC1394_COLOR_CODING_MONO8:
	  if (fmt != VIAM_HWFMT_MONO8)	continue;
	  break;
	case DC1394_COLOR_CODING_YUV411:
	  if (fmt != VIAM_HWFMT_YUV411)	continue;
	  break;
	case DC1394_COLOR_CODING_YUV422:
	  if (fmt != VIAM_HWFMT_YUV422_UYVY)	continue;
	  break;
	case DC1394_COLOR_CODING_YUV444:
	  if (fmt != VIAM_HWFMT_YUV444)	continue;
	  break;
	case DC1394_COLOR_CODING_RGB8:
	  if (fmt != VIAM_HWFMT_RGB888)	continue;
	  break;
	case DC1394_COLOR_CODING_MONO16:
	  if (fmt != VIAM_HWFMT_MONO16)	continue;
	  break;
	case DC1394_COLOR_CODING_RAW8:
	  if (fmt != VIAM_HWFMT_MONO8)	continue;
	  break;
	case DC1394_COLOR_CODING_RAW16:
	  if (fmt != VIAM_HWFMT_MONO16)	continue;
	  break;

	default: continue;
      }
      break;
    }
    if (j>=f7_mode.color_codings.num) continue;
    *color = f7_mode.color_codings.codings[j];

    if (f7_mode.max_size_x*f7_mode.max_size_y < minw*minh) {
      minw = f7_mode.max_size_x;
      minh = f7_mode.max_size_y;
      *hwgeo = modes.modes[i];
    }
  }
  if (*hwgeo == -1) return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);

  return 0;
}


/* --- viamgen1394_hwfps --------------------------------------------------- */

static const struct {
  viam_hwfps_t sw;
  dc1394framerate_t hw;
  double fps;
} fpsmap[] = {
  { VIAM_HWFPS_1_875,	DC1394_FRAMERATE_1_875,	1.875},
  { VIAM_HWFPS_3_75,	DC1394_FRAMERATE_3_75,	3.75},
  { VIAM_HWFPS_7_5,	DC1394_FRAMERATE_7_5,	7.5},
  { VIAM_HWFPS_15,	DC1394_FRAMERATE_15,	15.},
  { VIAM_HWFPS_30,	DC1394_FRAMERATE_30,	30.},
  { VIAM_HWFPS_60,	DC1394_FRAMERATE_60,	60.},
  { VIAM_HWFPS_120,	DC1394_FRAMERATE_120,	120.},
  { VIAM_HWFPS_240,	DC1394_FRAMERATE_240,	240.},
};

int
viamgen1394_fps(viam_hwfps_t swfps, double *hwfps)
{
  int i;

  for(i=0; i<sizeof(fpsmap)/sizeof(fpsmap[0]); i++)
    if (fpsmap[i].sw == swfps) {
      *hwfps = fpsmap[i].fps;
      return 0;
    }
  return Eviamlib_UNSUPPORTED_HWMODE;
}

static int
viamgen1394_hwfps(double fps, dc1394framerate_t *hwfps)
{
  int i;

  for(i=0; i<sizeof(fpsmap)/sizeof(fpsmap[0]); i++)
    if (fpsmap[i].fps == fps) {
      *hwfps = fpsmap[i].hw;
      return 0;
    }
  return Eviamlib_UNSUPPORTED_HWMODE;
}


/* --- viamgen1394_hwtrigger ----------------------------------------------- */

static const struct {
  viam_hwtrigger_t sw;
  dc1394trigger_mode_t hw;
  dc1394trigger_polarity_t pol;
} triggermap[] = {
  { VIAM_HWTRIGGER_MODE0_LOW,  DC1394_TRIGGER_MODE_0, DC1394_TRIGGER_ACTIVE_LOW },
  { VIAM_HWTRIGGER_MODE0_HIGH, DC1394_TRIGGER_MODE_0, DC1394_TRIGGER_ACTIVE_HIGH },
  { VIAM_HWTRIGGER_MODE1_LOW,  DC1394_TRIGGER_MODE_1, DC1394_TRIGGER_ACTIVE_LOW },
  { VIAM_HWTRIGGER_MODE1_HIGH, DC1394_TRIGGER_MODE_1, DC1394_TRIGGER_ACTIVE_HIGH }
};

static int
viamgen1394_hwtrigger(viam_hwtrigger_t swtrig,
                      dc1394trigger_mode_t *hwtrig,
                      dc1394trigger_polarity_t *pol)
{
  int i;

  for(i=0; i<sizeof(triggermap)/sizeof(triggermap[0]); i++)
    if (triggermap[i].sw == swtrig) {
      *hwtrig = triggermap[i].hw;
      *pol = triggermap[i].pol;
      return 0;
    }
  return Eviamlib_UNSUPPORTED_HWMODE;
}
