/*
 * Copyright (c) 2011 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Apr  6 2011
 */


#include <stdio.h>
#include <string.h>
#include <err.h>
#include <errno.h>
#include <sys/time.h>

#include "viam/viam_acq_driver.h"
#include "dc1394_common.h"


/* --- local data ---------------------------------------------------- */

#define VIAMLB_NUNITS	(6)

typedef struct viamlb_unit_t {
  int unit;				/**< unit number */
  struct viamlb_camera_t *master;	/**< master camera */
  char *buffer;				/**< image buffer for RGB888 */
} *viamlb_unit_t;

static struct viamlb_camera_t {
  uint64_t uid;				/**< firewire id */
  uint32_t attached;			/**< units in use */
  viam_onoff_t xmit;			/**< transmitting images */
  uint32_t capturing;			/**< DMA image buffer owned */
  viam_hwmode_t hwmode;			/**< current camera mode */

  viamgen1394_camera_t camera;
  struct viamlb_unit_t unit[VIAMLB_NUNITS];
} *lbcamera;
static int ncameras;

static int	viamlb_hardware_list(viam_handle_t vh);
static int	viamlb_hardware_attach(viam_handle_t vh,
			viam_camera_t camera);
static void	viamlb_hardware_detach(viam_handle_t vh,
			viam_camera_t camera);
static int	viamlb_camera_configure(viam_handle_t vh,
			viam_camera_t camera);
static int	viamlb_camera_unconfigure(viam_handle_t vh,
			viam_camera_t camera);
static int	viamlb_datatransmit(viam_handle_t vh, viam_camera_t camera,
			viam_onoff_t on);
static int	viamlb_dataflush(viam_handle_t vh, viam_camera_t camera);
static int	viamlb_capture(viam_handle_t vh, viam_camera_t camera);
static char *	viamlb_capturedata(viam_handle_t vh, viam_camera_t camera,
			size_t *bytes, size_t *rowstride);
static struct timeval
		viamlb_capturetimestamp(viam_handle_t vh,
			viam_camera_t camera);
static int	viamlb_capturedone(viam_handle_t vh, viam_camera_t camera);
static int	viamlb_queryhwmodes(viam_handle_t vh, viam_camera_t camera,
			int offset, viam_hwmode_t *supported, size_t n);
static int	viamlb_queryhwfeature(viam_handle_t vh, viam_camera_t camera,
			viam_filtertype_t filter, unsigned int *methods,
			unsigned int *aumodes);
static int	viamlb_sethwfeaturemode(viam_handle_t vh,
			viam_camera_t camera, viam_filtertype_t filter,
			viam_filtermethod_t method,
			viam_filterautomode_t aumode);
static int	viamlb_sethwfeaturevalue(viam_handle_t vh,
			viam_camera_t camera, viam_filtertype_t filter,
			const viam_filterargs_obj args);

static void	viamlb_camerainfo(dc1394camera_t *camera);

const struct viam_acqdev_obj viam_acqdev_entry = {
  "ladybug",
  viamgen1394_init,
  viamgen1394_fini,
  viamlb_hardware_list,
  viamlb_hardware_attach,
  viamlb_hardware_detach,
  viamlb_camera_configure,
  viamlb_camera_unconfigure,
  viamlb_datatransmit,
  viamlb_dataflush,
  viamlb_capture,
  viamlb_capturedata,
  viamlb_capturetimestamp,
  viamlb_capturedone,
  viamlb_queryhwmodes,
  viamlb_queryhwfeature,
  viamlb_sethwfeaturemode,
  viamlb_sethwfeaturevalue
};


/* --- viamlb_hardware_list ------------------------------------------------ */

/** \brief Print available camera devices on all buses
 *
 * \param[in] vh	Library handle.
 *
 * \return 0 on success.
 */

static int
viamlb_hardware_list(viam_handle_t vh)
{
  return viamgen1394_hardware_list(vh, viamlb_camerainfo);
}


/* --- viamlb_hardware_attach ---------------------------------------------- */

/** \brief Allocate camera ports and handles
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle to attach.
 *
 * \return 0 on success, or an error.
 */

static int
viamlb_hardware_attach(viam_handle_t vh, viam_camera_t camera)
{
  char *end, *end2;
  uint64_t uid;
  uint32_t min, max;
  dc1394error_t s;
  long unit = 0;
  int i, j;

  /* split uid into 64bits:unit integer */
  uid = strtoll(viam_camera_uid(camera), &end, 0);
  if (end == viam_camera_uid(camera) || (*end != ':' && *end != '\0')) {
    return viam_seterrno(vh, Eviamlib_BAD_ID);
  }

  /* parse unit */
  if (*end != '\0') {
    unit = strtol(++end, &end2, 0);
    if (end2 == end || unit < 0 || unit >= VIAMLB_NUNITS) {
      warnx("bad unit for camera %s", viam_camera_uid(camera));
      return viam_seterrno(vh, Eviamlib_BAD_ID);
    }
  }

  /* look for that uid in existing cameras or create one */
  for(i=0; i<ncameras; i++) if (lbcamera[i].uid == uid) break;
  if (i>=ncameras) {
    lbcamera = realloc(lbcamera, (i+1) * sizeof(*lbcamera));
    if (!lbcamera) return viam_seterrno(vh, Eviamlib_MEMORY);
    ncameras = i+1;
    lbcamera[i].uid = uid;
    lbcamera[i].attached = 0;
    lbcamera[i].xmit = 0;
    lbcamera[i].capturing = 0;
    lbcamera[i].hwmode.size = VIAM_HWSZ_INVALID;
    for(j=0; j<VIAMLB_NUNITS; j++) {
      lbcamera[i].unit[j].unit = j;
      lbcamera[i].unit[j].master = &lbcamera[i];
      lbcamera[i].unit[j].buffer = NULL;
    }

    /* invoke generic attach for master unit */
    lbcamera[i].camera =
      viamgen1394_hardware_attach(vh, viam_camera_name(camera), uid);

    /* set a packet size >0 to workaround a problem with libdc1394 if
     * packet_size is 0 initially. 'configure' will then set this correctly. */
    s = dc1394_format7_get_packet_parameters(
      lbcamera[i].camera->handle, DC1394_VIDEO_MODE_FORMAT7_0, &min, &max);
    if (s == DC1394_SUCCESS) {
      s = dc1394_format7_set_packet_size(
        lbcamera[i].camera->handle, DC1394_VIDEO_MODE_FORMAT7_0, min);
    }
  }
  if (!lbcamera[i].camera) return viam_seterrno(vh, Eviamlib_NO_SUCH_CAMERA);

  lbcamera[i].attached |= 1<<unit;
  viam_camera_sethwdata(camera, &lbcamera[i].unit[unit]);
  return 0;
}


/* --- viamlb_hardware_detach -------------------------------------- */

/** \brief Cleanup dclb handle
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 */

static void
viamlb_hardware_detach(viam_handle_t vh, viam_camera_t camera)
{
  viamlb_unit_t c = viam_camera_hwdata(camera);

  if (c) {
    if (c->master) {
      c->master->attached &= ~(1<<c->unit);
      if (!c->master->attached) {
        viamgen1394_hardware_detach(vh, c->master->camera);
      }
    }
    viam_camera_sethwdata(camera, NULL);
  }
}


/* --- viamlb_camera_configure ------------------------------------- */

/** \brief Setup camera with current configuration parameters.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 *
 * \return 0 on success
 */

static int
viamlb_camera_configure(viam_handle_t vh, viam_camera_t camera)
{
  viamlb_unit_t c = viam_camera_hwdata(camera);
  const viam_hwmode_t *swmode;
  double fps;
  int w, h;
  int s;

  if (!c) return viam_seterrno(vh, Eviamlib_NO_SUCH_CAMERA);
  if (!(swmode = viam_camera_hwmode(vh, camera)))
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);

  /* check params */
  if (swmode->crop != VIAM_HW_FIXED)
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
  if (swmode->trigger != VIAM_HWTRIGGER_INTERNAL)
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
  if (viam_hwmode_size(vh, swmode, &w, &h, NULL, NULL))
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
  if (swmode->fps == VIAM_HWFPS_FREE)
    fps = 0.;
  else if (viamgen1394_fps(swmode->fps, &fps))
    return viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);

  /* alloc interlacing buffer */
  if (swmode->format == VIAM_HWFMT_RGB888) {
    c->buffer = malloc(w*h*3);
    if (!c->buffer) return viam_seterrno(vh, Eviamlib_MEMORY);
  }

  /* already configured? */
  if (swmode->size == c->master->hwmode.size &&
      swmode->format == c->master->hwmode.format &&
      swmode->crop == c->master->hwmode.crop &&
      swmode->fps == c->master->hwmode.fps &&
      swmode->trigger == c->master->hwmode.trigger)
    return 0;

  /* map given format to ladybug format7 */
  h *= VIAMLB_NUNITS;
  if (swmode->format == VIAM_HWFMT_RGB888) h *= 4;
  s = viamgen1394_camera_configure(
    vh, c->master->camera,
    VIAM_HW_CROP, w, h, VIAM_HWFMT_MONO8, fps, swmode->trigger);
  if (s) return s;

  c->master->hwmode = *swmode;
  return 0;
}


/* --- viamlb_camera_unconfigure ----------------------------------- */

/** \brief Reset bank configuration
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 *
 * \return 0 on success
 */

static int
viamlb_camera_unconfigure(viam_handle_t vh, viam_camera_t camera)
{
  viamlb_unit_t c = viam_camera_hwdata(camera);
  int s;

  if (!c) return 0;
  s = viamgen1394_camera_unconfigure(vh, c->master->camera);
  if (c->buffer) free(c->buffer);
  return 0;
}


/* --- viam_datatransmit --------------------------------------------- */

int
viamlb_datatransmit(viam_handle_t vh, viam_camera_t camera, viam_onoff_t on)
{
  viamlb_unit_t c = viam_camera_hwdata(camera);
  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  if (c->master->xmit ^ on) {
    return viamgen1394_datatransmit(vh, c->master->camera, on);
    c->master->xmit = on;
  }
  return 0;
}


/* --- viam_dataflush ------------------------------------------------ */

static int
viamlb_dataflush(viam_handle_t vh, viam_camera_t camera)
{
  viamlb_unit_t c = viam_camera_hwdata(camera);
  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  return viamgen1394_dataflush(vh, c->master->camera);
}


/* --- viamlb_capture ---------------------------------------------- */

static int
viamlb_capture(viam_handle_t vh, viam_camera_t camera)
{
  viamlb_unit_t c = viam_camera_hwdata(camera);
  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  if (c->master->capturing++) return 0;

  return viamgen1394_capture(vh, c->master->camera);
}


/* --- viamlb_capturedata ------------------------------------------ */

static char *
viamlb_capturedata(viam_handle_t vh, viam_camera_t camera,
		     size_t *bytes, size_t *rowstride)
{
  viamlb_unit_t c = viam_camera_hwdata(camera);
  char *top;

  if (!c) { viam_seterrno(vh, Eviamlib_NOT_CONFIGURED); return NULL; }

  top = viamgen1394_capturedata(vh, c->master->camera, bytes, rowstride);
  if (!top) return NULL;

  switch(c->master->hwmode.size) {
    case VIAM_HWSZ_1024x768:
      return top + c->unit * 768 * *rowstride;

    case VIAM_HWSZ_512x384: {
      int i;
      unsigned char *d = (unsigned char*)c->buffer;
      unsigned char *r = (unsigned char*)top + 4*c->unit * 384 * *rowstride;
      unsigned char *g1 = r + 384 * *rowstride;
      unsigned char *g2 = g1 + 384 * *rowstride;
      unsigned char *b = g2 + 384 * *rowstride;

      for(i=384 * *rowstride; i>0; i--) {
        *d++ = *r++;
        *d++ = ((*g1^*g2)>>1) + (*g1&*g2); g1++; g2++;
        *d++ = *b++;
      }

      *rowstride *= 3;
      return c->buffer;
    }

    default: break;
  }

  viam_seterrno(vh, Eviamlib_UNSUPPORTED_HWMODE);
  return NULL;
}


/* --- viamlb_capturetimestamp ------------------------------------- */

static struct timeval
viamlb_capturetimestamp(viam_handle_t vh, viam_camera_t camera)
{
  viamlb_unit_t c = viam_camera_hwdata(camera);
  if (!c) {
    viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);
    return (struct timeval){0};
  }

  return viamgen1394_capturetimestamp(vh, c->master->camera);
}


/* --- viamlb_capturedone ------------------------------------------ */

static int
viamlb_capturedone(viam_handle_t vh, viam_camera_t camera)
{
  viamlb_unit_t c = viam_camera_hwdata(camera);
  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  if (--c->master->capturing) return 0;

  return viamgen1394_capturedone(vh, c->master->camera);
}


/* --- viamlb_queryhwmodes ------------------------------------------------- */

static int
viamlb_queryhwmodes(viam_handle_t vh, viam_camera_t camera, int offset,
		    viam_hwmode_t *supported, size_t n)
{
  viam_hwmode_t swmode;
  int count = 0;

  /* available formats are hardcoded to what format7 supports. If you want
   * other formats than format7, use the generic dc1394 driver */

  /* no crop */
  swmode.crop = VIAM_HW_FIXED;
  /* fps is free */
  swmode.fps = VIAM_HWFPS_FREE;
  /* internal trigger only */
  swmode.trigger = VIAM_HWTRIGGER_INTERNAL;

  /* 1024x768 MONO8 */
  swmode.size = VIAM_HWSZ_1024x768;
  swmode.format = VIAM_HWFMT_MONO8;
  supported[count++] = swmode;

  /* 512x384 RGB888 */
  swmode.size = VIAM_HWSZ_512x384;
  swmode.format = VIAM_HWFMT_RGB888;
  supported[count++] = swmode;

  return count;
}


/* --- viamlb_queryhwfeature --------------------------------------- */

static int
viamlb_queryhwfeature(viam_handle_t vh,
			viam_camera_t camera, viam_filtertype_t filter,
			unsigned int *methods, unsigned int *aumodes)
{
  viamlb_unit_t c = viam_camera_hwdata(camera);
  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  return viamgen1394_queryhwfeature(
    vh, c->master->camera, filter, methods, aumodes);
}


/* --- viamlb_sethwfeaturemode ------------------------------------- */

/** \brief Set hardware feature mode for a camera.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 * \param[in] filter	Feature type.
 * \param[in] method	Hardware mode.
 * \param[in] aumode	Auto mode.
 *
 * \return 0 on success.
 */

static int
viamlb_sethwfeaturemode(viam_handle_t vh, viam_camera_t camera,
			  viam_filtertype_t filter, viam_filtermethod_t method,
			  viam_filterautomode_t aumode)
{
  viamlb_unit_t c = viam_camera_hwdata(camera);
  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  return viamgen1394_sethwfeaturemode(
    vh, c->master->camera, filter, method, aumode);
}


/* --- viamlb_sethwfeaturevalue ------------------------------------ */

/** \brief Set harware feature value for a camera.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera handle.
 * \param[in] filter	Feature type.
 * \param[in] args	Feature value.
 *
 * \return 0 on success.
 */

static int
viamlb_sethwfeaturevalue(viam_handle_t vh, viam_camera_t camera,
			   viam_filtertype_t filter,
			   const viam_filterargs_obj args)
{
  viamlb_unit_t c = viam_camera_hwdata(camera);
  if (!c) return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  return viamgen1394_sethwfeaturevalue(vh, c->master->camera, filter, args);
}


/* --- viamlb_camerainfo ------------------------------------------- */

static void
viamlb_camerainfo(dc1394camera_t *camera)
{
  int i;

  if (camera->vendor_id != 0xb09d) return;
  /* apparently, model_id is not initialized */

  for(i=0;i<VIAMLB_NUNITS;i++) {
    printf("[node %s:0x%016llx:%d] %s (%s)\n",
           viam_acqdev_entry.name,
           camera->guid, i, camera->model, camera->vendor);
  }
}
